module CA (Rule , fromRule, ruleLength, CA, makeRule, sim, mkRandomRule,
  mkAllWhiteRule, mkAllBlackRule, mkRandomCA, ruleBits, caBits, caWidth,
  mkWhiteCA, simM)

where

import Ring
import System.Random (randomRIO)
import Control.Monad (replicateM, forM, foldM)
import Control.Monad.ST.Strict (unsafeIOToST)
import qualified Data.Foldable as F (foldl')
import Data.List (foldl')
import Data.Word (Word8)
import Data.Array.Base (unsafeAt, unsafeWrite, unsafeRead)
import Data.Array.Unboxed (UArray)
import qualified Data.Array.Unboxed as UArray (UArray, listArray)
import Data.Array.IArray (elems)
import GHC.ST (ST)
import qualified Data.Array.ST as ST (runSTUArray, newListArray, writeArray, newArray_, STUArray)
import qualified Data.Array.IO as IO (castIOUArray, newArray_, writeArray, readArray, IOUArray)
import Debug.Trace (trace)

data Rule = Rule { ruleWidth :: Int, rules :: UArray Int Bool }

instance Show Rule where
  show (Rule w r) = "[" ++ concat (map show (elems r)) ++ "]"
  
type CA = Ring 

-- ^ Create a rule, assuming the given bit width. The width specifies the number
-- of bits to either side of a cell in which to look. A width of 2 implies
-- that 5 bits will be examined, so 32 rules are need. A width of 3 implies
-- 7 bits, requiring 128 rules.
-- Therefore the association list should have at least 2 ^ (width * 2 + 1) mappings from integers
-- to bit values.
--
-- When a rule is applied, the automata cells are converted
-- to an integer value, which is used to look up the corresponding color in
-- the rule table.
makeRule :: Int -> [Bool] -> Rule
makeRule w rules = Rule w (ST.runSTUArray (ST.newListArray (0, (2 ^ (w * 2 + 1) - 1)) rules))

-- ^ Create a rule with new bits from an existing rule.
fromRule :: Rule -> [Bool] -> Rule
fromRule (Rule w _) = makeRule w

-- ^ Gets the bit patterns for this rule. The bits are arranged in ascending
-- order according to the value of the keys given. The pattern for each
-- bit is therefore implicit in the position of the bit.
ruleBits :: Rule -> [Bool]
ruleBits = elems . rules

-- ^ Number of bits that make up this rule. 
ruleLength :: Rule -> Int
ruleLength (Rule w _) = 2 ^ (w * 2 + 1)

-- ^ Width of the automata.
caWidth :: CA -> Int
caWidth = size

-- ^ Bits that make up an automata.
caBits :: CA -> [Bool]
caBits = toListR

-- ^ Advance the CA one step, using the
-- rule given. The automata is examined from left to right. The same
-- cell which was current when step began will be current when the step
-- ends.
step :: Int -> Rule -> CA -> CA
step rowLen rule@(Rule width rules) row =
  -- Apply rule to each cell in CA. 
  let getRule currIdx = {-# SCC "step_7" #-} pattern' start 0
        where
          start = (currIdx - width)
          end = (currIdx + width)
          pattern' cnt !result
            | cnt > end = result
            | otherwise = if (currR cnt row)
                            then pattern' (cnt + 1) (result * 2 + 1)
                            else pattern' (cnt + 1) (result * 2)
      cells = {-# SCC "step_5" #-} map (unsafeAt rules . getRule) [0 .. rowLen - 1]
      result = {-# SCC "step_8" #-} fromUArray (UArray.listArray (0, rowLen - 1) cells)
  in -- trace ("With rule: " ++ show (zip [0..] (ruleBits rule)) ++ "\n" ++ show row ++ "\n" ++ show result) 
         result

stepWithUArray :: Int -> Rule -> CA -> CA
stepWithUArray rowLen rule@(Rule width rules) row =
  -- Apply rule to each cell in CA. 
  let upper = {-# SCC "step_1" #-} rowLen - 1
      getRule currIdx = {-# SCC "step_7" #-} pattern' start 0
        where
          start = {-# SCC "step_2" #-} (currIdx - width)
          end = {-# SCC "step_3" #-} (currIdx + width)
          pattern' cnt !result
            | {-# SCC "step_4" #-} cnt > end = result
            | otherwise = {-# SCC "step_5" #-} if ({-# SCC "step_6" #-} currR cnt row)
                            then pattern' (cnt + 1) (result * 2 + 1)
                            else pattern' (cnt + 1) (result * 2)
      makeNewRow :: ST s (ST.STUArray s Int Bool)
      makeNewRow = {-# SCC "step_9" #-}
        do
          arr <- {-# SCC "step_10" #-} ST.newArray_ (0, upper)
          let fill idx
                | {-# SCC "step_11" #-} idx > upper = return ()
                | otherwise = {-# SCC "step_12" #-} do
                    {-# SCC "step_13" #-} unsafeWrite arr idx (unsafeAt rules (getRule idx))
                    {-# SCC "step_14" #-} fill (idx + 1)
          {-# SCC "step_15" #-} fill 0
          {-# SCC "step_6" #-} return $! arr
  in {-# SCC "step_8" #-} fromUArray (ST.runSTUArray makeNewRow)

stepWithFold :: Rule -> CA -> CA
stepWithFold rule@(Rule width rules) row =
  -- Apply rule to each cell in CA. 
  let newCell currIdx val = {-# SCC "step_6" #-}  unsafeAt rules ruleIdx
        where
          ruleIdx = {-# SCC "step_7" #-} pattern' start 0
            where
              start = (currIdx - width)
              end = (currIdx + width)
              pattern' cnt !result
                | cnt > end = result
                | otherwise = if (currR cnt row)
                                then pattern' (cnt + 1) (result * 2 + 1)
                                else pattern' (cnt + 1) (result * 2)
  in {-# SCC "step_8" #-} Ring.foldl1' newCell row

stepM :: Rule -> CA -> IO CA
stepM (Rule width rules) row =
  do
    let (lowerBits, upperBits) = (0, width * 2)
        (lowerRow, upperRow) = (0, size row - 1)
        bitIdxs = [lowerBits .. upperBits]
        rowIdxs = [0 .. upperRow]
    bits <- IO.newArray_ (lowerBits, upperBits) :: IO (IO.IOUArray Int Bool)
    let writeNeighbors currIdx = do
          let vals = map (\amt -> currR amt row) [(currIdx - width) .. (currIdx + width)] 
          mapM_ (uncurry (IO.writeArray bits)) (zip bitIdxs vals)
        applyRule :: IO Bool
        applyRule = do
          pattern <- (foldM toIntA 0 bitIdxs :: IO Int)
          return $! unsafeAt rules pattern
        toIntA result idx = do
          val <- IO.readArray bits idx
          return $!
            if val 
              then result * 2 + 1
              else result * 2
        newRow = ST.runSTUArray
          (do
            rowArr <- ST.newArray_ (0, upperRow) 
            let writeCell idx = do
                  unsafeIOToST $! writeNeighbors idx
                  cellVal <- unsafeIOToST applyRule 
                  unsafeWrite rowArr idx cellVal
            sequence_ $! map writeCell rowIdxs
            return rowArr)
    return $! fromUArray newRow
              

simM :: Int -> Rule -> CA -> IO (CA)
simM cnt rule initial = foldM (\row i -> stepM rule row) initial [1..cnt] 
      
-- ^ Simulate the CA infinitely, using the rule given. Return the final CA. Note that rules
-- are matched left to right. That is, when neighboring cells are examined
-- to match a rule, they are read left to right and the most significant
-- bit is on the left.
sim :: Rule -> CA -> [CA]
sim rule initial = iterate (stepWithUArray (caWidth initial) rule) initial

-- ^ Make a random CA row with the given width.
mkRandomCA :: Int -> IO CA
mkRandomCA width  =
  do
    vals <- replicateM width (randomRIO (0 :: Int, 1) >>= \n -> return $ if n == (0 :: Int) then False else True)
    return $! fromListR vals

mkWhiteCA :: Int -> IO CA
mkWhiteCA width = return $! (fromListR (replicate width False))

-- ^ Makes a random rule with the given number of neighbors on each sdie of the
-- cell. The rule has full coverage of all possible configurations. 
mkRandomRule :: Int -> IO Rule
mkRandomRule neighbors =
  do
    rules <- replicateM (2 ^ (neighbors * 2 + 1)) (randomRIO (0 :: Int, 1) >>= \n -> return $ if n == (0 :: Int) then False else True)
    return $! makeRule neighbors rules

mkAllWhiteRule :: Int -> Rule
mkAllWhiteRule w = makeRule w (repeat False)

mkAllBlackRule :: Int -> Rule
mkAllBlackRule w = makeRule w (repeat True)

-- mkUnchangedRule = makeRule 2 [(0,Zero),(1,Zero),(2,One),(3,One),(4,Zero),(5,Zero),(6,One),(7,One)]

main =
  do
    initial <- mkRandomCA 8
    rule <- mkRandomRule 1
    putStrLn $ show (zip [0..] (ruleBits rule))
    result <- simM 10 rule initial 
    (putStrLn . show) result
    -- mapM_ (putStrLn . show) (take 2 $ iterate (step rule) initial)
        