module Main

where

import CA
import Control.Monad (replicateM, unless)
import System.Random (randomRIO)
import Data.List (sortBy)
import Data.Array.IArray (Array,listArray, (!))
import Data.Time (getCurrentTime, diffUTCTime, UTCTime)
import Debug.Trace (trace)

type Fitness = Float

main = do
    let numRules = 25
        numGenerations = 2
        ruleWidth = 3 -- Makes a 7 bit rule.
        numCAs = 25
        caWidth = 149
        mutateProbability = 2
        crossOver = singlePointCrossOver 
        mutator = mutateWithProb mutateProbability
        iterations = caWidth * 2
        breedingPopulation = 9
        minFitness = 0.8
        classifier = classifyMajority
        ruleMaker = (mkRandomRule ruleWidth)
        caMaker = (replicateM numCAs (mkRandomCA caWidth))
    start <- getCurrentTime
    putStrLn $ "Starting at: " ++ show start
    putStrLn $ show numGenerations ++ " generations, " ++
                show numRules ++ " " ++ show (ruleWidth * 2 + 1) ++
                " bit rules each, " ++ show numCAs ++ " initial CAs per rule."
    rules <- evolve numRules numGenerations iterations breedingPopulation
      classifier
      minFitness
      start
      ruleMaker
      caMaker
      crossOver
      mutator
    putStrLn "Top 10 rules"
    mapM_ (putStrLn . show) (take 10 rules)
    end <- getCurrentTime
    putStrLn $ "Finished at: " ++ show end ++ ". Execution took " ++ (show $ diffUTCTime end start)
        
evolve :: Int -> Int -> Int -> Int -> (CA -> CA -> Int) -> Float -> UTCTime -> IO Rule -> IO [CA]
  -> (Rule -> Rule -> IO Rule) -> (Rule -> IO Rule) -> IO [(Rule, Fitness)]
evolve ruleCnt numGen iterations breedingPop classifier minFitness startTime mkRule mkCAs crossover mutator = evolve' startTime 1 []
  where
    evolve' start cnt prevRules =
          do
            putStrLn $ "Generation " ++ show cnt 
            -- Sort results by their fitness
            results <- tournament iterations ruleCnt classifier prevRules mkRule mkCAs
                        >>= return . filter ((>= minFitness) . snd)
            putStrLn $ "Top 10 of " ++ show (length results) ++ " that met fitness."
            mapM_ (\(r, f) -> putStrLn $ show f ++ ": " ++ show r) (take 10 results)
            if cnt == numGen
              then return $! results
              else
                do
                  -- mapM_ (putStrLn . show) (take 10 results)
                  newRules <- breed results breedingPop crossover mutator
                  -- putStrLn "New rules"; mapM (putStrLn . show) newRules;
                  end <- getCurrentTime
                  putStrLn $ "Finished generation " ++ show cnt ++ " at: " ++ show end ++ ". Execution took " ++ (show $ diffUTCTime end start)
                  evolve' end (cnt + 1) newRules 
        where
          fitness (_, f1) (_, f2) = compare f2 f1 

-- ^ If all cells are black, success. Otherwise, failure.
allBlack :: CA -> CA -> Int
allBlack _ final = if and (caBits final) then 1 else 0

alwaysWin :: CA -> CA -> Int
alwaysWin _ _  = 1

alwaysLose :: CA -> CA -> Int
alwaysLose _ _  = 0

-- ^ Takes initial and final CA row. Returns 1 if
-- the CA was classified correctly, zero otherwise.
classifyMajority :: CA -> CA -> Int
classifyMajority init final = 
  let (numInitWhite, numInitBlack) = (rowLength - numInitBlack, length (filter id (caBits init)))
      (numFinalWhite, numFinalBlack) = (rowLength - numFinalBlack, length (filter id (caBits final)))
      rowLength = caWidth init
      -- Return number of white (Zero) and black (One) cells in the row. Rowlength is a
      -- constant during the round, so number of black cells is the number of
      -- cells that weren't white.
  in if numInitWhite > numInitBlack && numFinalWhite == rowLength ||
          numInitBlack > numInitWhite && numFinalBlack == rowLength
      then 1
      else 0

noCrossOver :: Rule -> Rule -> IO Rule
noCrossOver rule1 = return . (const rule1)

noMutation :: a -> IO a
noMutation = return . id

-- Mutate each bit according to probabity given. Argument should
-- be between 0 and 100.
mutateWithProb :: Int -> Rule -> IO Rule
mutateWithProb prob rule =
    do
      newBits <- mapM mutate' (ruleBits rule)
      let newRule = fromRule rule newBits
      -- putStrLn $ "Mutated " ++ show rule ++ " to " ++ show newRule
      return $! newRule
  where
    mutate' b =
      do
        i <- randomRIO (0, 99)
        return $! if i < prob
          then (not b)
          else b
          
singlePointCrossOver :: Rule -> Rule -> IO Rule
singlePointCrossOver rule1 rule2 =
  do
    let ruleLen = ruleLength rule1 
    idx <- randomRIO (0, ruleLen - 1)
    let b1 = ruleBits rule1
        b2 = ruleBits rule2
        newRule = fromRule rule1 (take idx b1 ++ (drop idx b2))
    --putStrLn $ "Crossing " ++ show rule1 ++ " and " ++ show rule2 ++ " at " ++ show idx ++ " to get " ++ show newRule
    return $! newRule
  
-- Breed a list of rules based on the crossover and mutator
-- functions given.
breed :: [(Rule, Fitness)] -> Int -> (Rule -> Rule -> IO Rule) -> (Rule -> IO Rule) -> IO [Rule]
breed rules breedingPop crossover mutator =
  do
    -- Each rule has 9 partners to choose from, so generate a bunch of random numbers.
    if numRules > 1
      then do
            partnerIdxs <- replicateM (breedingPop * numRules) $ randomRIO (0, numRules - 1)
            children <- mapM breed' (zip3 rules (takeLists breedingPop partnerIdxs) [0..])
            mapM mutator children
      else
        mapM (mutator . fst) rules
  where
    ruleArr :: Array Int (Rule, Fitness)
    ruleArr = listArray (0, numRules - 1) rules 
    numRules = length rules
    -- Probabilistically select a mate for each rule and breed. Sometimes
    -- a rule may not breed and returns unchanged.
    takeLists :: Int -> [a] -> [[a]]
    takeLists _ [] = [[]]
    takeLists cnt ls = take cnt ls : (takeLists cnt (drop cnt ls))
    breed' ((rule, _), mateIndices, selfIdx) = do
      -- Rule can't cross with itself - so track index. Need to be
      -- careful at boundaries of list.
      let crossIdx idx 
            | idx == selfIdx && idx == numRules - 1 = idx - 1
            | idx == selfIdx = idx + 1
            | otherwise = idx
          -- Sort potential mates by their fitness
          mates = (sortBy fitness . map ((ruleArr !) . crossIdx)) mateIndices
          fitnessRange = sum (map snd mates)
      -- Pick a random number to select mate by
      p <- randomRIO (0, fitnessRange)
      -- putStrLn $ "Mates for " ++ show rule ++ " are " ++ show mates
      let mate = (dropWhile ((p >=) . snd) . drop 1 . scanl (\(_, fit) m -> ([fst m], fit + (snd m))) ([], 0)) mates
      if (null mate) 
        then return rule
        else crossover rule (head $ fst $ head $ mate)
    -- lowest to highest
    fitness (_, f1) (_, f2) = compare f1 f2
    
-- Run each rule on the automatas given. Return a fitness score for each
tournament :: Int -> Int -> (CA -> CA -> Int) -> [Rule] -> (IO Rule) -> (IO [CA]) -> IO [(Rule, Fitness)]
tournament iterations cnt classifier prevRules mkRule mkCAs = do
    -- Make initial CAs
    initial <- mkCAs
    -- Make random rules.
    rules <- replicateM (cnt - length prevRules) mkRule >>= return . (prevRules ++)
    return $! (zip rules (map (caRound iterations classifier initial) rules))
    
-- Determine how many automatas the rule is able to correctly classify. The
-- value returned is the number of correctly classified automata out of
-- all automata given.
caRound :: Int -> (CA -> CA -> Int) -> [CA] -> Rule -> Fitness
caRound iterations classifier initials rule = -- trace ("Top: " ++ show top ++ ", bot " ++ show bot ++ ", total " ++ show total)
    total
  where
    bot = (fromIntegral $ length initials)
    top = (fromIntegral $ sum (map (uncurry classifier) results))
    total = (top / bot)
    results = map (\initRow -> (initRow, (head . drop (iterations + 1) . sim rule) initRow)) initials
